
(define gen-cont 
     (lambda () (gensym 'c)))

(define gen-var 
     (lambda () (gensym 'v)))

(mkmac (form-type form)
       (if (or (atom? form) 
               (eq? (car form) 'quote))
           'atomic
           (car form)))

(define optimize? t)

(define cps 
   (lambda (form)
      (cps-form form '(lambda (x) x))))

(define cps-form
   (lambda (form cont)
       ((case (form-type form)
              (atomic cps-atom)
              (if cps-if)
              (call/cc cps-call/cc)
              (set! cps-set!)
              (begin cps-begin)
              (lambda cps-lambda)
              (else cps-combination))
           form 
           cont)))

(define cps-atom 
   (lambda (form cont) `(,cont ,form)))


(mkmac (predicate form)
       (cadr form))
(mkmac (consequent form)
       (caddr form))
(mkmac (alternative form)
       (cadddr form))

(define cps-if
   (lambda (form cont)
      (cps-form (predicate form)
                `(lambda (pred-val)
                    (if pred-val
                        ,(cps-form (consequent form) cont)
                        ,(cps-form (alternative form) cont))))))


(mkmac (lambda-vars form)
       (cadr form))
(mkmac (lambda-body form)
       (cddr form))
              
(define cps-lambda 
   (lambda (form cont)
      (let ((lambda-cont (gen-cont))
            (body (lambda-body form)))
         (let ((body (case (length body)
                        (0 nil)
                        (1 (car body))
                        (else `(begin ,@body)))))
            `(,cont (lambda (,@(lambda-vars form) ,lambda-cont)
                        ,(cps-form body lambda-cont)))))))


(mkmac (call/cc-body form)
       (cadr form))

(define cps-call/cc
   (lambda (form cont)
      (cps-form (call/cc-body form)
                (let ((f (gen-var))
                      (v (gen-var))
                      (k (gen-cont)))
                   `(lambda (,f)
                       (,f (lambda (,v ,k)
                             (,cont ,v))
                          ,cont))))))

(mkmac (set!-val form)
       (caddr form))
(mkmac (set!-var form)
       (cadr form))

(define cps-set! 
   (lambda (form cont)
       (let ((var (gen-var)))
         (cps-form (set!-val form)
                   `(lambda (,var)
                       (,cont (set! ,(set!-var form) ,var)))))))

(mkmac (begin-body forms)
       (cdr forms))
(mkmac (next-forms forms)
       (cdr forms))
(mkmac (first-form forms)
       (car forms))

(define cps-begin 
   (lambda (form cont)
      (letrec ((begin-forms 
                 (lambda (forms)
                    (if (null? (next-forms forms))
			(cps-form (first-form forms) cont)
                        `(lambda (dummy)
                            ,(cps-form (first-form forms)
                                       (begin-forms (next-forms forms))))))))
           (begin-forms (begin-body form)))))


(mkmac (combination-fcn form)
       (car form))
(mkmac (combination-args form)
       (cdr form))
(import 'getl)

(define combination-type
   (lambda (fcn)
      (cond ((atom? fcn)
             (case (car (getl fcn
                              '(scheme-global system-function
                                scheme-macro scheme-primitive)))
                (scheme-global 'combination)
                (system-function 'primop)
                (scheme-primitive 'primop)
                (scheme-macro 'macro)
                (else 'combination)))
          ((eq? (car fcn) 'lambda) 'lambda-combination)
          (t 'fcombination))))

(define cps-combination
   (lambda (form cont)
      (let ((fcn (combination-fcn form))
	    (args (combination-args form)))
	   (case (combination-type fcn)
		 (primop 
        	    (cps-args args
		       (lambda (vars)
			  `(,cont (,fcn ,@vars)))))
		 (macro (cps-form (expand form) cont))
		 (lambda-combination
		    (if optimize?
		       (cps-args args
			  (lambda (vars)
			     (let ((ncont (gen-cont)))
				  `((lambda (,@(lambda-vars fcn)
					       ,ncont)
				       ,(cps-form
					   `(begin ,@(lambda-body fcn))
					   ncont))
				    ,@vars
				    ,cont))))
		       (cps-args args
			  (lambda (vars)
			     (cps-form fcn
				`(lambda (fvar)
				    (fvar ,@vars ,cont)))))))
		 (combination
		    (if optimize? 
		       (cps-args args
			  (lambda (vars)
			     `(,fcn ,@vars ,cont)))			     
		       (cps-args args
			  (lambda (vars)
			     (cps-form fcn
				`(lambda (fvar)
				    (fvar ,@vars ,cont)))))))
		 (fcombination
		    (cps-args args
		       (lambda (vars)
			  (cps-form fcn
			     `(lambda (fvar)
				 (fvar ,@vars ,cont))))))
		 (else (warn "bad combination type " fcn))))))


(define cps-args 
   (lambda (args end-function)
       (letrec 
         ((do-args 
             (lambda (args vars)
                (cond ((null? args)
                       (end-function (reverse vars)))
                      ((and (atom? (car args)) optimize?)
                       (do-args (next-forms args)
                                (cons (car args) vars)))
                      (t (let ((var (gen-var)))
                            (cps-form (first-form args)
                                      `(lambda (,var)
                                          ,(do-args (next-forms args)
                                                    (cons var vars))))))))))
          (do-args args nil))))


(define expand
   (lambda (form)
      ((global-binding (combination-fcn form)) form)))

(define lisp-expand 
    (lambda (form macro)
        (warn "lisp macro " macro)))


(define map-file 
   (lambda (fcn file)
      (let ((ifile (open file 'read)))
	   (fluid-let ((input-port ifile))
	      ((rec loop 
		  (lambda ()
		     (let ((item (read)))
			  (if (not (equal? item eof))
			      (begin (fcn item)
				     (loop)))))))))))


;;; test data and misc fcns


(define ex1 '(call/cc
               (lambda (k)
                  (letrec ((f (lambda (n)
                                 (if (= 0 n) 
                                     '1
                                     (+ n (f (sub1 n)))))))
                       (call/cc (lambda (k)
                                   (cons (k (f 3)) (k (f 4)))))))))
;; value is 11
(define exf '(call/cc 
              (lambda (k)
                 (call/cc (lambda (k)
                             (cons (k (f 3)) (k (f 4))))))))



(define f 
   (lambda (val cont)
      (letrec ((ff (lambda (n)
                      (if (= 0 n)
                          '1
                          (+ n (ff (sub1 n)))))))
        (cont (ff val)))))
               

(define a '(f (a b) c))

(define b '((lambda (x y) (+ x y)) (a b) c))

(import '$prpr)
(import 'tyo)

(define doit
   (lambda (form)
      (let ((perty (lambda (f)
		      (writeln)
		      ($prpr f)
		      (writeln)
		      (writeln)
		      (writeln)))
	    (page (lambda ()
		     (tyo 12)
		     (writeln)
		     (writeln))))
	   (writeln "original form")
	   (perty form)
	   (writeln "original form compiled")
	   (perty (compile form))
	   (page)
	   (let ((cc (cps form))) 
		(writeln "cps form")
		(perty cc)
		(page)
		(writeln "cps form compiled")
		(perty (compile cc))
		(page)))))

(define run (lambda (e) (execute (compile (cps e)))))

