(declare (special scheme-directory 
                  **fnv**
		  **ticks** 
		  *toplevel-continuation*
		  *scheme-version*
		  *banner-line*
		  unprintable-symbols))

(def if
   (macro (l)
	  (cond ((eq (length l) 3) `(cond (,(cadr l) ,(caddr l))))
		((eq (length l) 4) `(cond (,(cadr l) ,(caddr l))
					  (t ,(cadddr l)))))))

(sstatus gcstrings t)  ; garbage collect strings

(declare (special *initial-env* *initial-args* *initial-result* *init-base*
	    piport poport *initial-pc* *initial-cont* development?))

;;; for system debugging setq development? to a non-nil value and
;;; a lisp debugger will be entered (instead of a (reset)) on
;;; vsm, compiler, etc. errors.

(setq development? nil)

(if (not (null development?))
      (progn (*rset t)			; these are needed for the 
	     (sstatus translink nil)))  ; debugger to work right

(def lisp-debug
   (lambda ()
      (if (not (null development?))
	  (debug '|developmental debugging|)
	  (reset))))



;;; Initial vsm register values

(declare (special *initial-fnv* keyboard))

(load (concat scheme-directory "base.o"))  ; base environment 
                                           ; functions
(setq *init-base* (genbase))

(setq *initial-fnv*
      (mapcar 'cons 
	 '(scheme-top-level input-port output-port  scheme-prompt 
	     unprintables scheme-macro-note scheme-primitive-note
	     system-function-note scheme-global-note
	     defined-forms defining-forms base-environment
	     system-base)
	 (list
	    nil				   ;;; scheme-top-level
	    (setq keyboard `(&port ,piport))  ;;; input-port
	    `(&port ,poport 74 . 1)           ;;; output-port
	    ">>> "	         		   ;;; scheme-prompt
	    '((&engine . <ENGINE>)		   ;;; unprintables
	      (&state . <STATE>)
	      (&closure . <CLOSURE>)
	      (&rest-closure . <CLOSURE>)
	      (&transform . <TRANSFORM>)
	      (&sys . <SYSTEM-FUNCTION>)
	      (&port . <PORT>)
	      (&base . <BASE-ENV>)
	      (&vector . <VECTOR>)
	      (&ref . <REF>)
	      (&unassigned-constant . <UNASSIGNED-CONSTANT>)
	      (&closed-port . <CLOSED-PORT>)
	      (&cont . <CONTINUATION>)
	      (&fluid-env . <FLUID-ENVIRONMENT>)
	      (&lexical-env . <LEXICAL-ENVIRONMENT>))    
	    t 			;;; scheme-macro-note
	    t			;;; scheme-primitive-note
	    t			;;; system-function-note
	    nil			;;; scheme-global-note
	    nil			;;; defined-forms
	    nil			;;; defining-forms
	    *init-base*         ;;; user base environment
	    *init-base*         ;;; system base environment
	    )))

(setq *initial-env* nil)

; all the other hand S-coded definitions are at the head of std.s and sys.s

(setq *initial-pc* '((*ha)))

(setq *toplevel-function*
   '(&closure (0 nil (*co . scheme-top-level) (*pr &ms . fluid1) (*ti . 0))))

(setq *initial-args* nil)

(setq *initial-result* nil)

(setq *initial-cont* nil)

(setq *toplevel-continuation*
   `((,*initial-pc*
	,*initial-args*
	.
	,*initial-env*)
     . 
     ,*initial-cont*))

(def run
   (lambda (pc)
      (progn (setq **fnv** *initial-fnv*)
	     (vsm pc 
		  *toplevel-continuation*
		  *initial-env*
		  *initial-args*
		  *initial-result*))))
   
(def load-scheme
   (lambda ()
      (prog (port)
	    (load (concat scheme-directory 'except.o)) ; exception handler
	    (load (concat scheme-directory 'error.o)) ; error handler
	    (load (concat scheme-directory 'print.o)) ; print stuff
	    (load (concat scheme-directory 'hostmacs.o)) ; host macros
	    (setq **fnv** *initial-fnv*)
	    (load (concat scheme-directory 'env.o))   ; initial environment
	    ; must be loaded before
	    ; env!!
	    (load (concat scheme-directory 'constant.o)) ; constants
	    (load (concat scheme-directory 'vector.o)); vectors
;	    (load (concat scheme-directory 'prims.o)) ; primitives
	    (load (concat scheme-directory 'prims.l)) ; primitives
	    (load (concat scheme-directory 'vsm.o))   ; virtual machine
	    (load (concat scheme-directory 'comp.o))  ; scheme compiler in 
	    ; lisp
	    (load (concat scheme-directory 'expand.o)); beta-expander
	    (load (concat scheme-directory 'mkmac.l)) ; mkmac
	    (load (concat scheme-directory 'signal.o)); signal handlers
;	    (load (concat scheme-directory 'debug.o)); history debugger
;	    (load (concat scheme-directory 'debug.l)); history debugger
	    (setq port (infile (concat scheme-directory 'sys.s)))
	    (run (compile (read port)))
	    (close port)
	    )))

(setq user-top-level
    '(lambda ()
	(prog ()
	    loop
	    (princ *banner-line*)
	    (princ *scheme-version*)
	    (terpri)
	    (setq **fnv** *initial-fnv*)
	    (setq **ticks** nil)
	    (errset (vsm '((*ti . 0)) 
			 *initial-cont*
			 *initial-env*
			 *initial-args*
			 *toplevel-function*))
	    (go loop))))

(setq unprintable-symbols 
   (mapcar 'car (cdr (assq 'unprintables *initial-fnv*))))

(setq 			; initialize virtual machine registers
      **fnv** nil
      **ticks** nil)

(load-scheme)
