; Virtual Scheme Machine for Scheme 84

;;;
;;; To compile for frequency statistics incant:
;;; "liszt -e '(sstatus feature freq)' comp.l"
;;;

; patch for interpretation only:

(setq #+freq t)
(def i-count 
     (nlambda (x) t))


;#+freq
;(eval-when (load)
;   (load "freq"))

;#+freq 
;(eval-when (compile)
;   (princ "Compiling VSM for frequency analysis")
;   (terpri)
;   (load "freq")
;   ;; for instructions and i-count macro
;   (apply (function special) instructions))

(declare (special =env= =class= =temp= =temp1= **args**
	    unprintable-symbols **comp-env**))

; Main Registers

(declare (special **ticks**
		  **engine-cont**
		  **engine-failure**
		  **engine-success**
                  **fnv**))

; Debugging registers

(declare (special *error* namespacetypes *error-data* *error-res*))





(def if
   (macro (l)
	  (cond ((eq (length l) 3) `(cond (,(cadr l) ,(caddr l))))
		((eq (length l) 4) `(cond (,(cadr l) ,(caddr l))
					  (t ,(cadddr l)))))))

(def mk-continuation
   (macro (l)
	  `(cons (cons ,(cadr l)
			(cons ,(caddr l) (cons ,(cadddr l) ,(caddddr l))))
		 ,(cadddddr l))))

(def &frame (macro (l) `(car ,(cadr l))))

(def &pc (macro (l) `(car ,(cadr l))))

(def &args (macro (l) `(cadr ,(cadr l))))

(def &env (macro (l) `(caddr ,(cadr l))))

(def &fnv (macro (l) `(cdddr ,(cadr l))))

(def &cont (macro (l) `(cdr ,(cadr l))))

(def top-stack (macro (l) '(car **args**)))

(def 2nd-stack (macro (l) '(cadr **args**)))

(def 3rd-stack (macro (l) '(caddr **args**)))

; S-code interpreter

(def vsm
  (lambda (**pc** **cont** **env** **args** **result**)
    (prog (=opcode= =frame= =funtype=  =n= =rib-loc=
	   =pair= =op= =opargs= =inst= =rib= =rib-depth= =offset=)
        (go redirected-pc-loop)     
      loop 
        (setq **pc** (cdr **pc**))
      redirected-pc-loop
        (and **ticks**
	     (cond ((zerop (setq **ticks** (1- **ticks**)))
		    (setq **args** 
		       (list (cons '&engine
				(cons (mk-continuation **pc**
				       **args** **env** **fnv** **cont**)
				      **result**))))
		    (setq **result** **engine-failure**)
		    (setq =opargs= 1)
	            (go engine-return))))
	(setq =inst= (car **pc**))
	(setq =opcode= (car =inst=))
	
	(cond
	  ((eq =opcode= '*pu) #+freq (i-count *pu)
	   (go *push))
	  ((eq =opcode= '*lr) #+freq (i-count *lr)
	   (go *local-lookup))
	  ((eq =opcode= '*pr) #+freq (i-count *pr)
	   (go *primitive))
	  ((eq =opcode= '*at) #+freq (i-count *at)
	   (go *application-tail))
	  ((eq =opcode= '*ti) #+freq (i-count *ti)
	   (go *thunk-invocation))
	  ((eq =opcode= '*co) #+freq (i-count *co)
	   (go *constant))
	  ((eq =opcode= '*if) #+freq (i-count *if)
	   (go *if-tail))
	  ((eq =opcode= '*le) #+freq (i-count *le)
	   (go *let-lambda))
	  ((eq =opcode= '*sa) #+freq (i-count *sa)
	   (go *save))
	  ((eq =opcode= '*re) #+freq (i-count *re)
	   (go *restore))
	  ((eq =opcode= '*la) #+freq (i-count *la)
	   (go *lambda))
	  ((eq =opcode= '*gr) #+freq (i-count *gr)
	   (go *global-lookup))
	  ((eq =opcode= '*ls) #+freq (i-count *ls)
	   (go *local-set))
	  ((eq =opcode= '*aa) #+freq (i-count *aa)
	   (go *apply))
	  ((eq =opcode= '*rla) #+freq (i-count *rla)
	   (go *rest-lambda))
	  ((eq =opcode= '*rle) #+freq (i-count *rle)
	   (go *rest-let-lambda))
	  ((eq =opcode= '*ha) #+freq (i-count *ha)
	   (go *halt))
	  ((eq =opcode= '*ex) #+freq (i-count *ex)
	   (go *execute))
	  ((eq =opcode= '*gs) #+freq (i-count *gs)
	   (go *global-set))
	  ((eq =opcode= '*ra) #+freq (i-count *ra)
	   (go *result->args))
	  (t (go bad-vsm-opcode)))

 engine-return
    (setq **cont** **engine-cont**)
    (setq **ticks** nil)
    (go *application)

 *restore
    (cond ((null **cont**) (go process-ran-out)))
    (setq =frame= (&frame **cont**))
    (setq **cont** (&cont **cont**))
    (setq **pc** (&pc =frame=))
    (setq **args** (&args =frame=))
    (setq **env** (&env =frame=))
    (setq **fnv** (&fnv =frame=))
    (go redirected-pc-loop)

 *constant
    (setq =opargs= (cdr =inst=))
    (cond ((and (dtpr =opargs=) (eq (car =opargs=) '&unassigned-constant))
	   (setq **result** (lookup-constant (cdr =opargs=)))
	   (rplacd =inst= **result**))
	  (t (setq **result** =opargs=)))
    (go loop)

 *lambda
    (setq =opargs= (cdr =inst=))
    (setq **result** (cons '&closure (cons =opargs= **env**)))
    (go loop)

 *rest-lambda
    (setq =opargs= (cdr =inst=))
    (setq **result** (cons '&rest-closure (cons =opargs= **env**)))
    (go loop)
 
 *call/cc
   (setq **args** (list (cons '&cont **cont**)))
   (go loop)

 *engine
;   (setq =opargs= (cdr =inst=))
    (setq **result** (cons '&engine
			(cons (mk-continuation =opargs= nil
				 **env** **fnv** nil) nil)))
    (go loop)

 *if-tail
    (setq =opargs= (cdr =inst=))
    (cond (**result** (setq **pc** (car =opargs=)))
	  (t (setq **pc** (cdr =opargs=))))
    (go redirected-pc-loop)

 *result->args
    (setq **args** **result**)
    (go loop)

 *apply
    (cond ((null **args**) (go *thunk-invocation)))
    (cond ((not (dtpr **args**)) 
	   (raise (list 'SE%vsm 0 t '|Illegal argument to apply:| **args**)))
	  ((proc? **args**)
	   (raise (list 'SE%vsm 0 t '|Illegal argument to apply:| **args**)))
	  (t (setq **args** (properargs **args**))
	     (setq =opargs= (length **args**))))
    (go *application)

 *let-lambda
    (setq =opargs= (cdr =inst=))
    (setq **env** (cons (cons **args** =opargs=) **env**))
    (setq **args** nil)
    (go loop)

 *rest-let-lambda
    (setq =opargs= (cdr =inst=))
    (setq **env** 
       (cons (cons 
		(cons (nthcdr (length (cdr =opargs=)) **args**) **args**)
	        =opargs=)
       	     **env**))
    (go loop)

 *thunk-invocation
    (cond ((not (dtpr **result**)) (go bad-function)))
    (setq =funtype= (car **result**))
    (cond ((eq =funtype= '&closure) (go apply-thunk))
	  ((eq =funtype= '&rest-closure) (go apply-rest-thunk))
	  ((eq =funtype= '&state) (go start-state))
	  ((eq =funtype= '&sys) (go system-call))
	  ((eq =funtype= '&cont) (go apply-continuation))
	  ((eq =funtype= '&engine) (go start-engine))
	  ((eq =funtype= 'vector) (go apply-vector))
	  (t (go bad-function)))

 *application-tail
    (setq =opargs= (cdr =inst=))
 *application
    (cond ((not (dtpr **result**)) (go bad-function)))
    (setq =funtype= (car **result**))
    (cond ((eq =funtype= '&closure) (go apply-closure))
	  ((eq =funtype= '&rest-closure) (go apply-rest-closure))
	  ((eq =funtype= '&sys) (go system-call))
	  ((eq =funtype= '&cont) (go apply-continuation))
	  ((eq =funtype= '&vector) (go apply-vector))
	  ((eq =funtype= '&engine) (go start-engine))
	  ((eq =funtype= '&state) (go start-state))
	  (t (go bad-function)))
 
 start-state
    (cond (**args** (go wrong-number-of-args-to-state)))
    (setq =temp= (cdr **result**))
    (setq **result** (cdr =temp=))
    (setq =temp= (car =temp=))
    (setq =frame= (&frame =temp=))
    (setq **pc** (&pc =frame=))
    (setq **env** (&env =frame=))
    (setq **args** (&args =frame=))
    (setq **fnv** (&fnv =frame=))
    (setq **cont** (append (&cont =temp=) **cont**))
    (go redirected-pc-loop)

 start-engine
    (cond ((not (eq =opargs= 3)) (go wrong-number-of-args-to-engine))
	  (**ticks** (scherror "Sorry, can't nest engines") (go loop))
	  ((and (numberp (car **args**))
		(plusp (car **args**)))
	   (setq **ticks** (1+ (car **args**)))
	   (setq **engine-success** (cadr **args**))
           (setq **engine-failure** (caddr **args**))
	   (setq **engine-cont** **cont**)
	   (setq =temp= (cdr **result**))
	   (setq **result** (cdr =temp=))
	   (setq =temp= (car =temp=))
	   (setq =frame= (&frame =temp=))
	   (setq **pc** (&pc =frame=))
	   (setq **env** (&env =frame=))
	   (setq **args** (&args =frame=))
	   (setq **fnv** (&fnv =frame=))
	   (setq **cont** (&cont =temp=))
	   (go redirected-pc-loop))
	  ((zerop (car **args**)) 
	   (setq =opargs= 1)
	   (setq =temp= **result**)
	   (setq **result** (caddr **args**))
	   (setq **args** (list =temp=))
	   (setq **ticks** nil)
	   (go *application))
	  (t (setq **ticks** nil)
	     (raise (list 'SE%vsm 0 t '|illegal ticks for engine:|
		       (car **args**)))))

 apply-thunk
    (setq **result** (cdr **result**))
    (setq =n= (car (car **result**)))
    (cond  ((zerop =n=) (setq **env** (cdr **result**)))
	   (t (go wrong-number-of-args-to-closure)))
    (setq **pc** (cddr (car **result**)))
    (go redirected-pc-loop)

 apply-closure
    (setq **result** (cdr **result**))
    (setq =n= (car (car **result**)))
    (cond ((= =opargs= =n=)
	   (setq **env** (cons (cons **args** (cadr (car **result**)))
			       (cdr **result**))))
	  (t (go wrong-number-of-args-to-closure)))
    (setq **pc** (cddr (car **result**)))
    (setq **args** nil)
    (go redirected-pc-loop)

 apply-rest-thunk
    (setq **result** (cdr **result**))
    (setq =n= (car (car **result**)))
    (cond
       ((zerop =n=) 
	(setq **env** (cons (cons (cons nil nil) (cadr (car **result**)))
			    (cdr **result**))))
       (t (go wrong-number-of-args-to-closure)))
    (setq **pc** (cddr (car **result**)))
    (go redirected-pc-loop)

 apply-rest-closure
    (setq **result** (cdr **result**))
    (setq =n= (car (car **result**)))
    (cond ((or (lessp =n= =opargs=) (= =n= =opargs=))
	   (setq **env** (cons (cons (cons (nthcdr =n= **args**) **args**)
				     (cadr (car **result**)))
			       (cdr **result**))))
	  (t (go wrong-number-of-args-to-closure)))
    (setq **pc** (cddr (car **result**)))
    (setq **args** nil)
    (go redirected-pc-loop)
    

 system-call
    (setq **result** (apply (cdr **result**) **args**))
    (setq **args** nil)          
    (go *restore)  ;; sys fcns are called with *at 
                   ;; so they should do their own return's

 apply-vector
    (cond ((= =opargs= 1)
	   (setq **result** (vactor-ref **result** (car **args**)))
	   (setq **args** nil)
	   (go *restore))
	  (t (go wrong-number-args-to-vector)))
	    
 apply-continuation
    (cond ((= =opargs= 1)
	   (setq **cont** (cdr **result**))
	   (setq **result** (car **args**))
	   (go *restore))
	  (t (go wrong-number-of-args-to-continuation)))

 *push
    (setq **args** (cons **result** **args**))
    (go loop)

 *save
    (setq =opargs= (cdr =inst=))
    (setq **cont** (mk-continuation =opargs=
		      **args** **env** **fnv** **cont**))
    (setq **args** nil)
    (go loop)

 *halt
    (return **result**)

 *execute
    (setq **pc** **result**)
    (setq **env** nil)
    (go redirected-pc-loop)

 *local-set
    (setq =opargs= (cdr =inst=))
    (setq =rib-depth= (car =opargs=))
    (setq =offset= (cdr =opargs=))
;    #+freq
;    (lex-count =rib-depth= =offset=)
    (cond 
       ((= 0 =rib-depth=) (setq =rib= (caar **env**)) (go *offset-set))
       ((= 1 =rib-depth=) (setq =rib= (caadr **env**)) (go *offset-set))
       ((= 2 =rib-depth=) (setq =rib= (caaddr **env**)) (go *offset-set)))
    (setq =rib= (car (nthelem (1+ =rib-depth=) **env**)))
 *offset-set
    (cond
       ((= 0 =offset=) (rplaca =rib= **result**) (go loop))
       ((= 1 =offset=) (rplaca (cdr =rib=) **result**) (go loop))
       ((= 2 =offset=) (rplaca (cddr =rib=) **result**) (go loop)))
    (rplaca (nthcdr =offset= =rib=) **result**)
    (go loop)
    
 *global-set
    (setq =opargs= (cdr =inst=))
    (setq =pair= (baselocation =opargs=))
    (cond ((null =pair=)
	   (setq =temp= (global-namespacetype =opargs=))
	   (cond ((or (not =temp=)
		      (eq =temp= 'base-identifier)
		      (eq =temp= 'scheme-primitive)
		      (eq =temp= 'system-function))
		  (extendbase =opargs= **result**)
		  (cond ((memq =temp= '(scheme-primitive system-function))
			 (remprop =opargs= =temp=))))
		 ((and (eq =temp= 'constant)
		       (setq =pair= (getl =opargs=
				            '(scheme-constant
					      constant-primitive
					      constant-system-function)))
		       (eq (cadr =pair=) 'unassigned-constant))
		  (rplaca (cdr =pair=) **result**))
		 (t (setq *error* =opargs=)
		    (scherror
		       (concat =opargs= " already declared as a " =temp=)))))
          (t (rplacd =pair= **result**)
             (cond ((scheme-fluid 'scheme-global-note)
                    (princ "[Redefining ")
	            (print =opargs=)
	            (princ "]")
	            (terpri)))))
    (go loop)
    
 *local-lookup
    (setq =opargs= (cdr =inst=))
    (setq =rib-depth= (car =opargs=))
    (setq =offset= (cdr =opargs=))
;    #+freq
;    (lex-count =rib-depth= =offset=)
    (cond 
       ((= 0 =rib-depth=) (setq =rib= (caar **env**)) (go *offset))
       ((= 1 =rib-depth=) (setq =rib= (caadr **env**)) (go *offset))
       ((= 2 =rib-depth=) (setq =rib= (caaddr **env**)) (go *offset)))
    (setq =rib= (car (nthelem (1+ =rib-depth=) **env**)))
 *offset    
    (cond
       ((= 0 =offset=) (setq **result** (car =rib=)) (go loop))
       ((= 1 =offset=) (setq **result** (cadr =rib=)) (go loop))
       ((= 2 =offset=) (setq **result** (caddr =rib=)) (go loop)))
    (setq **result** (nthelem (1+ =offset=) =rib=))
    (go loop)
    
 *global-lookup
    (setq =opargs= (cdr =inst=))
    (setq =pair= (baselocation =opargs=))
    (cond ((null =pair=)
	   (setq *error-data* =opargs=)
	   (setq *error* **cont**)
	   (scherror "Unassigned identifier:"))
          (t (setq **result** (cdr =pair=))))
    (go loop)


 *primitive
    (setq =opargs= (cdr =inst=))
; Scheme primitives
; 
; Scheme primitives take their arguments from the top of **args**,
; popping them off and leaving the result in **result**

; When apply-primitive is called, the top two elements of the =opargs=
; are the class and name of the primitive

 *primitive-apply
      (setq =class= (car =opargs=))
      (setq =op= (cdr =opargs=))
      (cond ((eq =class= '&se) (go &select))
	    ((eq =class= '&ls) (go &list))
	    ((eq =class= '&pr) (go &predicate))
	    ((eq =class= '&tp) (go &type))
	    ((eq =class= '&no) (go &numeric))
	    ((eq =class= '&s4) (go &select-four))
	    ((eq =class= '&ve) (go &vector))
	    ((eq =class= '&np) (go &npred))
	    ((eq =class= '&io) (go &input/output))
	    ((eq =class= '&si) (go &system-interface))
	    ((eq =class= '&tr) (go &transcendental))
            ((eq =class= '&ms) (go &misc))
	    (t (go bad-class)))

 &select
    (cond ((eq =op= 'car) (go %car))
	  ((eq =op= 'cdr) (go %cdr))
	  ((eq =op= 'caaar) (go %caaar))
	  ((eq =op= 'caadr) (go %caadr))
	  ((eq =op= 'caar) (go %caar))
	  ((eq =op= 'cadar) (go %cadar))
	  ((eq =op= 'caddr) (go %caddr))
	  ((eq =op= 'cadr) (go %cadr))
	  ((eq =op= 'cdaar) (go %cdaar))
	  ((eq =op= 'cdadr) (go %cdadr))
	  ((eq =op= 'cdar) (go %cdar))
	  ((eq =op= 'cddar) (go %cddar))
	  ((eq =op= 'cdddr) (go %cdddr))
	  ((eq =op= 'cddr) (go %cddr)))

 &list
    (cond ((eq =op= 'cons) (go %cons))
	  ((eq =op= 'set-car!) (go %set-car!))
	  ((eq =op= 'set-cdr!) (go %set-cdr!))
	  ((eq =op= 'copy-no-constant) (go %copy-no-constant))
	  ((eq =op= 'delete) (go %delete))
	  ((eq =op= 'delq) (go %delq))
	  ((eq =op= 'last-pair) (go %last-pair))
	  ((eq =op= 'length) (go %length))
	  ((eq =op= 'member) (go %member))
	  ((eq =op= 'nth) (go %nth))
	  ((eq =op= 'list-ref) (go %list-ref))
	  ((eq =op= 'list-tail) (go %list-tail))
	  ((eq =op= 'reverse) (go %reverse))
	  ((eq =op= 'reverse!) (go %reverse!))
	  ((eq =op= 'transpose) (go %transpose)))

 &predicate
    (cond ((eq =op= 'null?) (go %null?))
	  ((eq =op= 'eq?) (go %eq?))
	  ((eq =op= 'eqv?) (go %eqv?))
	  ((eq =op= 'memv) (go %memv))
	  ((eq =op= 'assv) (go %assv))
	  ((eq =op= 'assoc) (go %assoc))
	  ((eq =op= 'assq) (go %assq))
	  ((eq =op= 'port?) (go %port?))
	  ((eq =op= 'equal?) (go %equal?))
	  ((eq =op= 'memq) (go %memq))
	  (t (go bad-op)))

 &type
    (cond ((eq =op= 'atom?) (go %atom?))
	  ((eq =op= 'number?) (go %number?))
	  ((eq =op= 'string?) (go %string?))
	  ((eq =op= 'proc?) (go %proc?))
	  ((eq =op= 'fix?) (go %fix?))
	  ((eq =op= 'float?) (go %float?))
	  ((eq =op= 'pair?) (go %pair?))
	  ((eq =op= 'ref?) (go %ref?))
	  ((eq =op= 'symbol?) (go %symbol?))
	  ((eq =op= 'syntactic-extension?) (go %syntactic-extension?))
	  ((eq =op= 'scheme-constant?) (go %scheme-constant?))
	  (t (go bad-op)))

 &numeric
    (cond ((eq =op= 'add1) (go %add1))
	  ((eq =op= 'sub1) (go %sub1))
	  ((eq =op= '+) (go %+))
	  ((eq =op= '*) (go %*))
	  ((eq =op= '-) (go %-))
	  ((eq =op= 'minus) (go %minus))
	  ((eq =op= '/) (go %/))
	  ((eq =op= 'abs) (go %abs))
	  ((eq =op= 'factorial) (go %factorial))
	  ((eq =op= 'fix) (go %fix))
	  ((eq =op= 'ceiling) (go %ceiling))
	  ((eq =op= 'truncate) (go %truncate))
	  ((eq =op= 'round) (go %round))
	  ((eq =op= 'float) (go %float))
	  ((eq =op= 'quotient) (go %quotient))
	  ((eq =op= 'remainder) (go %remainder))
	  ((eq =op= 'random) (go %random))
	  ((eq =op= 'sqrt) (go %sqrt))
	  (t (go bad-op)))

 &select-four
    (cond ((eq =op= 'caaaar) (go %caaaar))
	  ((eq =op= 'caaadr) (go %caaadr))
	  ((eq =op= 'caadar) (go %caadar))
	  ((eq =op= 'caaddr) (go %caaddr))
	  ((eq =op= 'cadaar) (go %cadaar))
	  ((eq =op= 'cadadr) (go %cadadr))
	  ((eq =op= 'caddar) (go %caddar))
	  ((eq =op= 'cadddr) (go %cadddr))
	  ((eq =op= 'cdaaar) (go %cdaaar))
	  ((eq =op= 'cdaadr) (go %cdaadr))
	  ((eq =op= 'cdadar) (go %cdadar))
	  ((eq =op= 'cdaddr) (go %cdaddr))
	  ((eq =op= 'cddaar) (go %cddaar))
	  ((eq =op= 'cddadr) (go %cddadr))
	  ((eq =op= 'cdddar) (go %cdddar))
	  ((eq =op= 'cddddr) (go %cddddr)))
	  
 &vector
    (cond ((eq =op= 'vector-ref) (go %vector-ref))
	  ((eq =op= 'vector-set!) (go %vector-set!))
	  ((eq =op= 'primitive-make-vector) (go %primitive-make-vector))
	  ((eq =op= 'vector-length) (go %vector-length))
	  ((eq =op= 'vector->list) (go %vector->list))
	  ((eq =op= 'list->vector) (go %list->vector))
	  ((eq =op= 'vector-fill!) (go %vector-fill!))
	  ((eq =op= 'vector?) (go %vector?)))

 &npred
    (cond ((eq =op= '=0) (go %=0))
	  ((eq =op= '=) (go %=))
	  ((eq =op= '<) (go %<))
	  ((eq =op= '>) (go %>))
	  ((eq =op= '>0) (go %>0))
	  ((eq =op= '<0) (go %<0))
	  ((eq =op= '<=) (go %<=))
	  ((eq =op= '>=) (go %>=))
	  ((eq =op= 'alpha<) (go %alpha<))
	  (t (go bad-op)))
 
 &input/output
    (cond  ((eq =op= 'close) (go %close))
	   ((eq =op= 'current-column) (go %current-column))
	   ((eq =op= 'display) (go %display))
	   ((eq =op= 'display&) (go %display&))
	   ((eq =op= 'file-exists?) (go %file-exists?))
	   ((eq =op= 'flush-input) (go %flush-input))
	   ((eq =op= 'flush-output) (go %flush-output))
	   ((eq =op= 'line-length) (go %line-length))
	   ((eq =op= 'newline) (go %newline))
	   ((eq =op= 'open) (go %open))
	   ((eq =op= 'print) (go %print))
	   ((eq =op= 'print-length) (go %print-length))
	   ((eq =op= 'print&) (go %print&))
	   ((eq =op= 'prompt-read) (go %prompt-read))
	   ((eq =op= 'read) (go %read))
	   ((eq =op= 'read-atom) (go %read-atom))
	   ((eq =op= 'read-char) (go %read-char))
	   ((eq =op= 'read/port) (go %read/port))
	   ((eq =op= 'read-atom/port) (go %read-atom/port))
	   ((eq =op= 'read-char/port) (go %read-char/port))
	   ((eq =op= 'set-line-length!) (go %set-line-length!)))

 &system-interface
    (cond  ((eq =op= 'pretty) (go %pretty))
	   ((eq =op= 'compile) (go %compile))
	   ((eq =op= 'mkmac-match?) (go %mkmac-match?))
	   ((eq =op= 'add-to-syntax-table) (go %add-to-syntax-table))
	   ((eq =op= 'beta-expand) (go %beta-expand))
	   ((eq =op= 'expand-once) (go %expand-once))
	   ((eq =op= 'pp-exp) (go %pp-exp))
	   ((eq =op= 'copying-intern*) (go %copying-intern*))
	   ((eq =op= 'beta-tag) (go %beta-tag))
	   ((eq =op= 'exit) (go %exit))
	   ((eq =op= 'reset) (go %reset))
	   ((eq =op= 'declare-constant) (go %declare-constant))
           ((eq =op= 'undeclare-constant) (go %undeclare-constant))
	   ((eq =op= 'function-alias) (go %function-alias))
	   ((eq =op= 'gc) (go %gc))
	   ((eq =op= 'top-level-ids) (go %top-level-ids))
	   ((eq =op= 'scheme-constants) (go %scheme-constants))
	   ((eq =op= 'beta-transforms) (go %beta-transforms))
	   ((eq =op= 'base-identifiers) (go %base-identifiers))
	   ((eq =op= 'scheme-primitives) (go %scheme-primitives))
           ((eq =op= 'system-functions) (go %system-functions))
	   ((eq =op= 'reify) (go %reify))
	   ((eq =op= 'global-namespace-type) (go %global-namespace-type))
	   ((eq =op= 'import) (go %import))
	   ((eq =op= 'lisp-eval) (go %lisp-eval))
	   ((eq =op= 'make-printable) (go %make-printable))
	   ((eq =op= 'make-unprintable) (go %make-unprintable))
	   ((eq =op= 'ptime) (go %ptime))
	   ((eq =op= 'remove-from-namespace) (go %remove-from-namespace))
	   ((eq =op= 'scheme-reset) (go %scheme-reset))
	   ((eq =op= 'transcript-off) (go %transcript-off))
	   ((eq =op= 'transcript-on) (go %transcript-on))
	   ((eq =op= 'add-mkmac-name) (go %add-mkmac-name))
	   ((eq =op= 'remove-mkmac-name) (go %remove-mkmac-name))
	   ((eq =op= 'set-lexical-semantics) (go %set-lexical-semantics))
	   ((eq =op= 'set-application-semantics)
	    (go %set-application-semantics))
	   ((eq =op= 'set-literal-semantics) (go %set-literal-semantics))
	   ((eq =op= 'scoped?) (go %scoped?))
	   (t (go bad-op)))

 &transcendental
     (cond  ((eq =op= 'arccos) (go %arccos))
	    ((eq =op= 'arcsin) (go %arcsin))
	    ((eq =op= 'arctan) (go %arctan)) ; 1-Mar-85
	    ((eq =op= 'atan) (go %atan))
	    ((eq =op= 'cos) (go %cos))
	    ((eq =op= 'exp) (go %exp))
	    ((eq =op= 'expt) (go %expt))
	    ((eq =op= 'log) (go %log))
	    ((eq =op= 'sin) (go %sin))
	    ((eq =op= 'tan) (go %tan))
	    (t (go bad-op)))

 &misc
    (cond ((eq =op= 'engine-state) (go %engine-state))
	  ((eq =op= 'call/cc) (go %call/cc))
	  ((eq =op= 'prim-execute) (go %prim-execute))
	  ((eq =op= 'engine1) (go %engine1))
	  ((eq =op= 'primitive-apply) (go %primitive-apply))
          ((eq =op= 'fluid:bound?) (go %fluid:bound?))
	  ((eq =op= 'fluid1) (go %fluid1))
	  ((eq =op= 'set-fluid1) (go %set-fluid1))
	  ((eq =op= 'let-fluid1) (go %let-fluid1))
	  ((eq =op= 'ascii->symbol) (go %ascii->symbol))
	  ((eq =op= 'symbol->ascii) (go %symbol->ascii))
	  ((eq =op= 'deref) (go %deref))
	  ((eq =op= 'explode) (go %explode))
          ((eq =op= 'genbase) (go %genbase))
	  ((eq =op= 'gensym) (go %gensym))
	  ((eq =op= 'string->uninterned) (go %string->uninterned))
	  ((eq =op= 'getprop) (go %getprop))
	  ((eq =op= 'global-binding) (go %global-binding))
	  ((eq =op= 'implode) (go %implode))
	  ((eq =op= 'proplist) (go %proplist))
	  ((eq =op= 'putprop) (go %putprop))
	  ((eq =op= 'ref) (go %ref))
	  ((eq =op= 'remprop) (go %remprop))
	  ((eq =op= 'set-ref!) (go %set-ref!))
	  ((eq =op= 'subst) (go %subst))
	  ((eq =op= 'substring) (go %substring))
	  ((eq =op= 'swap-ref!) (go %swap-ref!))
	  (t (go bad-op)))

 %car (setq **result** (car **result**)) (go loop)
 %cdr (setq **result** (cdr **result**)) (go loop)
 %caaar (setq **result** (caaar **result**)) (go loop)
 %caadr (setq **result** (caadr **result**)) (go loop)
 %caar (setq **result** (caar **result**)) (go loop)
 %cadar (setq **result** (cadar **result**)) (go loop)
 %caddr (setq **result** (caddr **result**)) (go loop)
 %cadr (setq **result** (cadr **result**)) (go loop)
 %cdaar (setq **result** (cdaar **result**)) (go loop)
 %cdadr (setq **result** (cdadr **result**)) (go loop)
 %cdar (setq **result** (cdar **result**)) (go loop)
 %cddar (setq **result** (cddar **result**)) (go loop)
 %cdddr (setq **result** (cdddr **result**)) (go loop)
 %cddr (setq **result** (cddr **result**)) (go loop)

 %caaaar (setq **result** (caaaar **result**)) (go loop)
 %caaadr (setq **result** (caaadr **result**)) (go loop)
 %caadar (setq **result** (caadar **result**)) (go loop)
 %caaddr (setq **result** (caaddr **result**)) (go loop)
 %cadaar (setq **result** (cadaar **result**)) (go loop)
 %cadadr (setq **result** (cadadr **result**)) (go loop)
 %caddar (setq **result** (caddar **result**)) (go loop)
 %cadddr (setq **result** (cadddr **result**)) (go loop)
 %cdaaar (setq **result** (cdaaar **result**)) (go loop)
 %cdaadr (setq **result** (cdaadr **result**)) (go loop)
 %cdadar (setq **result** (cdadar **result**)) (go loop)
 %cdaddr (setq **result** (cdaddr **result**)) (go loop)
 %cddaar (setq **result** (cddaar **result**)) (go loop)
 %cddadr (setq **result** (cddadr **result**)) (go loop)
 %cdddar (setq **result** (cdddar **result**)) (go loop)
 %cddddr (setq **result** (cddddr **result**)) (go loop)
 
 %cons (setq **result** (cons **result** (top-stack))) (go pop1)
 %set-car! (cond
	      [(dtpr **result**)
	       (cond
		  [(proc? **result**)
		   (raise (list 
			     'SE%vsm
			      0 t '|Illegal argument to set-car! :|
			       **result**))]
	          [t (setq **result** (rplaca **result** (top-stack)))
                     (go pop1)])]
	      [t (raise (list 'SE%vsm 0 t 
			   '|Bad argument to set-car!:| **result**))])
 %set-cdr! (cond
	      [(dtpr **result**)
	       (cond
		  [(proc? **result**)
		   (raise (list 
			     'SE%vsm
			      0 t '|Illegal argument to set-cdr! :|
			       **result**))]
	          [t (setq **result** (rplacd **result** (top-stack)))
                     (go pop1)])]
	      [t (raise (list 'SE%vsm 0 t 
			   '|Bad argument to set-cdr! :| **result**))])

 %copy-no-constant (setq **result** (copy-no-constant **result**)) (go loop)
 %delete (setq **result** (delete **result** (top-stack))) (go pop1)
 %delq (setq **result** (delq **result** (top-stack))) (go pop1)
 %last-pair (setq **result** (last **result**)) (go loop)
 %length (setq **result** (length **result**)) (go loop)
 %member (setq **result** (member **result** (top-stack))) (go pop1)
 %nth (setq **result** (nthelem **result** (top-stack))) (go pop1)
 %list-ref (setq **result** (nthelem (1+ (top-stack)) **result**)) (go pop1)
 %list-tail (setq **result** (nthcdr (top-stack) **result**)) (go pop1)
 %reverse (setq **result** (reverse **result**)) (go loop)
 %reverse! (setq **result** (nreverse **result**)) (go loop)
 %transpose (setq **result** (transpose **result**)) (go loop)

 %null? (setq **result** (null **result**))
	(go loop)
 %eq? (setq **result** (eq **result** (top-stack)))
      (go pop1)
  %eqv? (setq **result** (eqv **result** (top-stack)))
       (go pop1)
 %memv (setq **result** (memv **result** (top-stack)))
       (go pop1)
 %assv (setq **result** (assv **result** (top-stack)))
       (go pop1)
 %assoc (setq **result** (assoc **result** (top-stack))) (go pop1)
 %assq (setq **result** (assq **result** (top-stack))) (go pop1)
 %port? (setq **result** (port? **result**)) (go loop)
 %equal? (setq **result** (equal **result** (top-stack))) (go pop1)
 %memq (setq **result** (memq **result** (top-stack))) (go pop1)

 %atom? (setq **result** (atom **result**))
	(go loop)
 %number? (setq **result** (numberp **result**))
	  (go loop)
 %string? (setq **result** (stringp **result**))
	  (go loop)
 %proc? (setq **result** (proc? **result**))
        (go loop)
 %fix? (setq **result** (fixp **result**)) (go loop)
 %float? (setq **result** (floatp **result**)) (go loop)
 %pair? (setq **result** (and (dtpr **result**)
			      (not (memq (car **result**)
					 unprintable-symbols))))
	    (go loop)
 %ref? (setq **result** (and (dtpr **result**)
			     (eq '&ref (car **result**)))) (go loop)
 %symbol? (setq **result** (symbolp **result**)) (go loop)
 %syntactic-extension? (setq **result** (syntactic-extension? **result**))
                       (go loop)
 %scheme-constant? (setq **result** (scheme-constant? **result**)) (go loop)

 %add1 (setq **result** (add1 **result**))
     (go loop)
 %sub1 (setq **result** (sub1 **result**))
     (go loop)
 %+ (setq **result** (plus **result** (top-stack)))
    (go pop1)
 %* (setq **result** (times **result** (top-stack)))
    (go pop1)
 %- (setq **result** (difference **result** (top-stack)))
    (go pop1)
 %minus (setq **result** (minus **result**))
     (go loop)
 %/ (setq **result** (quotient **result** (top-stack)))
    (go pop1)
 %abs (setq **result** (abs **result**)) (go loop)
 %factorial (setq **result** (fact **result**)) (go loop)
 %fix (setq **result** (fix **result**)) (go loop)
 %ceiling (cond ((fixp **result**))
		((zerop (difference **result** (fix **result**))))
		(t (setq **result** (fix (add1 **result**)))))
          (go loop)
 %truncate (cond ((fixp **result**))
		 (t (setq **result** (* (cond ((plusp **result**) 1)
					      (t -1))
					(fix (abs **result**))))))
           (go loop)
 %round (setq **result** (fix (plus **result** 0.5))) (go loop)
 %float (setq **result** (float **result**)) (go loop)
 %quotient (setq **result** (/ **result** (top-stack))) (go pop1)
 %remainder (setq **result** (mod **result** (top-stack))) (go pop1)
 %random (setq **result** (random **result**)) (go loop)
 %sqrt (setq **result** (sqrt **result**)) (go loop)

 %vector? (setq **result** (vactor? **result**)) (go loop)
 %list->vector (setq **result** (list->vactor **result**)) (go loop)
 %vector->list (setq **result** (vactor->list **result**)) (go loop)
 %vector-ref (setq **result** (vactor-ref **result** (top-stack))) (go pop1)
 %vector-set! (setq **result**
		 (vactor-set! **result** (top-stack) (2nd-stack)))
              (go pop2)
 %vector-length (setq **result** (vactor-length **result**)) (go loop)
 %vector-fill! (setq **result** (vactor-fill! **result** (top-stack)))
	       (go pop1)
 %primitive-make-vector (setq **result** (make-vactor **result**)) (go loop)

 %=0 (setq **result** (and (numberp **result**) (zerop **result**)))
     (go loop)
 %= (setq **result** (zerop (difference **result** (top-stack))))
    (go pop1)
 %< (setq **result** (lessp **result** (top-stack)))
    (go pop1)
 %> (setq **result** (greaterp **result** (top-stack)))
    (go pop1)
 %>0 (setq **result** (plusp **result**))
     (go loop)
 %<0 (setq **result** (minusp **result**))
     (go loop)
 %>= (setq **result** (not (lessp **result** (top-stack))))
     (go pop1)
 %<= (setq **result** (not (greaterp **result** (top-stack))))
     (go pop1)
 %alpha< (setq **result** (alphalessp **result** (top-stack))) 
         (go pop1)

 %close (setq **result** (schclose **result**)) (go loop)
 %current-column (setq **result** (current-column)) (go loop)
 %display (setq **result** (display **result**)) (go loop)
 %display& (setq **result** (display& **result**)) (go loop)
 %file-exists? (setq **result** (probef **result**)) (go loop)
 %flush-input (setq **result** nil)
              (drain (lport (scheme-fluid 'input-port))) (go loop)
 %flush-output (setq **result** nil)
               (drain (lport (scheme-fluid 'output-port))) (go loop)
 %line-length (setq **result** (line-length (scheme-fluid 'output-port)))
              (go loop)
 %newline (setq **result** (new-line)) (go loop)
 %open (setq **result** (open **result** (top-stack))) (go pop1)
 %print (setq **result** (schprint **result**)) (go loop)
 %print-length (setq **result** (pntlen **result**)) (go loop)
 %print& (setq **result** (schprint& **result**)) (go loop)
 %prompt-read (setq **result** (prompt-read **result**)) (go loop)
 %read (setq **result** (schread)) (go loop)
 %read-atom (setq **result** (read-atom)) (go loop)
 %read-char (setq **result** (read-char)) (go loop)
 %read/port (setq **result** (schread/port **result**)) (go loop)
 %read-atom/port (setq **result** (read-atom/port **result**)) (go loop)
 %read-char/port (setq **result** (read-char/port **result**)) (go loop)
 %set-line-length! (setq **result**
		      (set-line-length! (scheme-fluid 'output-port)
			                **result**))
                   (go loop)
 %pretty (setq **result** (pretty **result**)) (go loop)
 %compile (setq **result** (compile **result**)) (go loop)
 %mkmac-match? (setq **result**
		  (mkmac-match? **result** (top-stack) (2nd-stack)))
               (go pop2)
 %declare-constant
    (declare-constant **result**)
    (go loop)
 %undeclare-constant
    (undeclare-constant **result**)
    (go loop)
 %exit (setq **result** (exit)) (go loop) ; it never really gets to loop!
 %add-to-syntax-table (setq **result**
			 (add-to-syntax-table **result** (top-stack)))
                      (go pop1)
 %beta-expand (setq **result** (beta-expand **result**)) (go loop)
 %expand-once (setq **result** (expand-once **result**)) (go loop)
 %pp-exp ($prpr **result**)
         (setq **result** (terpri)) (go loop)
 %copying-intern* (setq **result** (copying-intern* **result**)) (go loop)
 %beta-tag (setq **result** (tag-frees **result** nil)) (go loop)
 %function-alias
    (setq **result** (function-alias **result**(top-stack))) (go pop1)
 %gc (setq **result** (gc)) (go loop)
 %top-level-ids (setq **result** (top-level-assigned-identifiers)) (go loop)
 %scheme-constants (setq **result** (symbols-bound-to-constants)) (go loop)
 %beta-transforms (setq **result** (beta-transforms)) (go loop)
 %base-identifiers (setq **result** (base-identifiers))  (go loop)
 %scheme-primitives (setq **result** (scheme-primitives)) (go loop)
 %system-functions (setq **result** (system-functions)) (go loop)
 %reify (setq **result** (reify **result**)) (go loop)
 %global-namespace-type
    (setq **result** (global-namespacetype **result**)) (go loop)
 %import (setq **result** (import **result**)) (go loop)
 %lisp-eval (setq **result** (eval **result**)) (go loop)
 %make-printable (setq **result** (make-printable **result**)) (go loop)
 %make-unprintable (setq **result** (make-unprintable **result**
				       (top-stack)))
                  (go pop1)
 %ptime (setq **result** (ptime)) (go loop)
 %remove-from-namespace
  (setq **result** (remove-from-namespace **result**)) (go loop)
 %reset (setq **result** (reset)) (go loop)
 %scheme-reset (setq **result** (scheme-reset)) (go loop)
 %transcript-off (setq **result** (transcript-off)) (go loop)
 %transcript-on (setq **result** (transcript-on **result**)) (go loop)
 %add-mkmac-name (setq **result** (add-mkmac-name **result**)) (go loop)
 %remove-mkmac-name (setq **result** (remove-mkmac-name **result**)) (go loop)
 %set-lexical-semantics (set-lexical-semantics **result**)
 			(setq **result** t) (go loop)
 %set-application-semantics (set-application-semantics **result**)
 			    (setq **result** t) (go loop)
 %set-literal-semantics (set-literal-semantics **result**)
                         (setq **result** t) (go loop)
 %scoped? (setq **result** (memq **result** (car **comp-env**))) (go loop)
 %arccos (setq **result** (acos **result**)) (go loop)
 %arcsin (setq **result** (asin **result**)) (go loop)
 %arctan (setq **result** (atan **result** 1)) (go loop) ; 1-Mar-85
 %atan (setq **result** (atan **result** (top-stack))) (go pop1)
	     
 %cos (setq **result** (cos **result**)) (go loop)
 %exp (setq **result** (exp **result**)) (go loop)
 %expt (setq **result** (expt **result** (top-stack))) (go pop1)
 %log (setq **result** (log **result**)) (go loop)
 %sin (setq **result** (sin **result**)) (go loop)
 %tan (setq **result** (quotient (sin **result**)
			  (cos **result**))) (go loop)

 %engine-state (cond ((eq (car **result**) '&engine)
                      (setq **result** (cons '&state (cdr **result**)))
		      (go loop))
		     (t (raise 
			   (list 'SE%vsm 0 t 
			      '|Bad argument to engine-state|))))
 %call/cc
    ;; we must reset **cont** in case call/cc function "falls through"
    ;; without invoking the current continuation!!
    (setq **cont**
       (mk-continuation (cdr **pc**) **args** **env** **fnv** **cont**))
    (setq **args** 
       (list (cons '&cont **cont**)))
    (setq =opargs= 1)
    (go *application)
 %prim-execute (go *execute)
 %engine1 (setq =opargs= (cddr (cadr **result**)))
          (setq **env** (cddr **result**))
	  (go *engine)
 %primitive-apply (setq **args** (top-stack))
	          (go *apply)  
 %fluid:bound? (setq **result**
		  (cond ((assq **result** **fnv**) t)))
               (go loop)
 %fluid1 (setq **result** (scheme-fluid **result**))
         (go loop)
 %set-fluid1
         (set-scheme-fluid **result** (top-stack))
	 (setq **result** (top-stack))
	 (go pop1)
 %let-fluid1
         (setq **fnv**
	     (nconc (mapcar 'cons **result** (top-stack)) **fnv**))
	 (setq **result** (2nd-stack))
	 (setq **args** (cddr **args**))
	 (go *thunk-invocation)
 %ascii->symbol (setq **result** (ascii->symbol **result**)) (go loop)
 %symbol->ascii (setq **result** (symbol->ascii **result**)) (go loop)
 %deref (setq **result** (cdr **result**)) (go loop)
 %explode (setq **result** (aexplodec **result**)) (go loop)
 %genbase (setq **result** (genbase)) (go loop)
 %gensym (setq **result** (gensym **result**)) (go loop)
 %string->uninterned (setq **result** (uconcat **result**)) (go loop)
 %getprop (setq **result** (get **result** (top-stack))) (go pop1)
 %global-binding (setq **result** (global-binding **result**)) (go loop)
 %implode (setq **result** (implode **result**)) (go loop)
 %proplist (setq **result** (plist **result**)) (go loop)
 %putprop (setq **result** 
	    (putprop **result** (top-stack) (2nd-stack))) 
         (go pop2)
 %ref (setq **result** (cons '&ref **result**)) (go loop)
 %remprop (setq **result** (remprop **result** (top-stack))) (go pop1)
 %set-ref! (setq **result** (rplacd **result** (top-stack))) (go pop1)
 %subst (setq **result** 
	  (subst **result** (top-stack) (2nd-stack))) 
       (go pop2)
 %substring (setq **result** 
	      (substring **result** (top-stack) (2nd-stack))) 
           (go pop2)
 %swap-ref! (setq =temp= (cdr **result**))
            (rplacd **result** (top-stack))
	    (setq **result** =temp=) (go pop1)
		     
 pop1 (setq **args** (cdr **args**))
      (go loop)

 pop2 (setq **args** (cddr **args**))
      (go loop)

 pop3 (setq **args** (cdddr **args**))
      (go loop)

 bad-class (setq *error* =class=)
           (scherror "Bad primitive class")
           (go loop)
 bad-op 
     (setq *error* =op=)
     (scherror "Bad primitive op")
     (go loop)

 

 bad-vsm-opcode
    (setq *error* =opcode=)
    (scherror "Bad vsm opcode")
    (go loop)

 process-ran-out
    (cond (**ticks** (setq **args** (list **result** (1- **ticks**)))
	             (setq **result** **engine-success**)
		     (setq =opargs= 2)
	             (go engine-return))
	  (t (scherror "Null continuation invoked, but no running engine")
	     (go loop)))

 bad-function
    (setq *error* **cont**)
    (setq *error-args* **args**)
    (setq *error-data* **result**)
    (scherror "Bad function")
    (go loop)

 wrong-number-of-args-to-closure
    (setq *error* **cont**)
    (setq *error-res* **result**)
    (setq *error-data* **args**)
    (scherror "Wrong number of arguments to closure")
    (go loop)

 wrong-number-of-args-to-engine
    (setq *error-res* **result**)
    (setq *error-data* **args**)
    (scherror "Wrong number of arguments to engine")
    (go loop)

 wrong-number-of-args-to-state
    (setq *error-res* **result**)
    (setq *error-data* **args**)
    (scherror "Wrong number of arguments to state")
    (go loop)

 wrong-number-args-to-vector
    (setq *error* **result**)
    (setq *error-data* **args**)
    (scherror "Wrong number of arguments to vector")
    (go loop)

 wrong-number-of-args-to-continuation
    (setq *error* **cont**)
    (setq *error-data* **args**)
    (scherror "Wrong number of arguments to continuation")
    (go loop))))

(def properargs
   (lambda (l)
      (cond
	 [(dtpr l) (cons (car l) (properargs (cdr l)))]
	 [(null l) nil]
	 [t (raise 
	       (list 'SE%vsm 0 t
		   '|Illegal argument to apply:| **args**))])))
